@@ -5,7 +5,7 @@ from typing import List, Set, Dict, Any, Type, Union, Tuple
 
 from sync.backends.inmemory import InMemoryLinkProvider
 from sync.datasource import DataSource
-from sync.mapping import Mapping, Mode, AttributeMap
+from sync.mapping import Mapping, Mode, AttributeMap, DirtyMode
 from sync.mixins import Timestamped, LinkProvider
 from sync.operations import (
     Addition,
@@ -14,6 +14,7 @@ from sync.operations import (
     AttributeUpdated,
     SkipUpdateOrphanedEntity,
     DeletedOrphanedEntity,
+    CouldNotCreate,
 )
 
 MODES_TO_PARAMETERS = {
@@ -94,10 +95,7 @@ class Syncer:
         operations = []
         try:
             conditions = self._create_key_value_pairs(
-                from_entity,
-                mapping.keys,
-                from_ds=from_ds,
-                to_ds=data_source,
+                from_entity, mapping.keys, from_ds=from_ds, to_ds=data_source,
             )
             found_entities = data_source.find(entity_type, conditions)
         except NotLinked as e:
@@ -109,16 +107,34 @@ class Syncer:
 
         if len(found_entities) == 0:
             from_entity_timestamp = self._timestamp(from_entity)
-            non_key_attributes = self._create_key_value_pairs(
-                from_entity,
-                mapping.attributes,
-                from_ds=from_ds,
-                to_ds=data_source,
-            )
-            to_entity = Syncer._create_entity(
-                conditions, non_key_attributes, entity_type, from_entity_timestamp
-            )
-            data_source.create(to_entity)
+
+            try:
+                non_key_attributes = self._create_key_value_pairs(
+                    from_entity, mapping.attributes, from_ds=from_ds, to_ds=data_source,
+                )
+                to_entity = Syncer._create_entity(
+                    conditions, non_key_attributes, entity_type, from_entity_timestamp
+                )
+            except NotLinked as e:
+                operations.append(
+                    SkipMissingLink(from_entity.__class__, entity_type, from_entity)
+                )
+                logging.warning(e)
+                return operations
+
+            if not mapping.create or not data_source.create(to_entity):
+                logging.warning(f"Could not add {to_entity.__repr__()}")
+                operations.append(
+                    CouldNotCreate(from_entity.__class__, entity_type, from_entity)
+                )
+                return operations
+
+            if mapping.dirty_mode == DirtyMode.UPDATE:
+                setattr(from_entity, mapping.dirty_attribute_name, False)
+                logging.info(
+                    f"Removed {mapping.dirty_attribute_name} flag from {from_entity}"
+                )
+                from_ds.persist(from_entity)
 
             operations.append(
                 Addition(from_entity.__class__, entity_type, from_entity, to_entity)
@@ -131,6 +147,7 @@ class Syncer:
                 operations.append(Linking(classes, entities))
         elif len(found_entities) == 1:
             to_entity = next(iter(found_entities))
+
             if self.link_provider.link(to_entity, from_entity, data_source, from_ds):
                 classes = frozenset(mapping.entity_types)
                 entities = frozenset({from_entity, to_entity})
@@ -177,6 +194,8 @@ class Syncer:
             if len(path) > 1:
                 key, *rest = path
                 next_entity = Syncer._resolve_entity(entity, key)
+                if not next_entity:
+                    return entities
                 entities = entities.union(
                     Syncer._collect_objects_in_path(next_entity, rest)
                 )
@@ -201,6 +220,16 @@ class Syncer:
             logging.debug(f"Setting timestamp of {entity} to {timestamp}")
             entity.update_timestamp(timestamp)
 
+    @staticmethod
+    def _is_dirty(mapping, entity):
+        try:
+            return getattr(entity, mapping.dirty_attribute_name)
+        except AttributeError as e:
+            logging.warning(
+                f"asked to read {mapping.dirty_attribute_name} from {entity} but no such attribute exists"
+            )
+            raise e
+
     def _update_linked_entity(
         self,
         from_ds: DataSource,
@@ -215,16 +244,10 @@ class Syncer:
         try:
             attributes = {
                 **self._create_key_value_pairs(
-                    from_entity,
-                    mapping.keys,
-                    from_ds=from_ds,
-                    to_ds=to_ds,
+                    from_entity, mapping.keys, from_ds=from_ds, to_ds=to_ds,
                 ),
                 **self._create_key_value_pairs(
-                    from_entity,
-                    mapping.attributes,
-                    from_ds=from_ds,
-                    to_ds=to_ds,
+                    from_entity, mapping.attributes, from_ds=from_ds, to_ds=to_ds,
                 ),
             }
         except NotLinked as e:
@@ -240,7 +263,7 @@ class Syncer:
 
             if not refreshed_to_entity:
                 logging.warning(
-                    f"Orphaned entities in link from {from_entity}: {refreshed_to_entity}"
+                    f"Orphaned entities in link from {from_entity}: {to_entity}"
                 )
                 operations.append(
                     SkipUpdateOrphanedEntity(
@@ -262,6 +285,11 @@ class Syncer:
             else:
                 to_entity = refreshed_to_entity
 
+            if mapping.dirty_mode == DirtyMode.DO_NOT_OVERWRITE and self._is_dirty(
+                mapping, to_entity
+            ):
+                continue
+
             # guard that to_entity has last been changed before or at the same time as from_entity
             from_entity_objects = Syncer._collect_objects_in_paths(
                 from_entity, mapping.from_paths()
@@ -297,14 +325,17 @@ class Syncer:
                 entity_updated = True
             if entity_updated:
                 to_ds.persist(to_entity)
+                if mapping.dirty_mode == DirtyMode.UPDATE:
+                    setattr(from_entity, mapping.dirty_attribute_name, False)
+                    logging.info(
+                        f"Removed {mapping.dirty_attribute_name} flag from {from_entity}"
+                    )
+                    from_ds.persist(from_entity)
 
         return operations
 
     def _sync_from_to(
-        self,
-        from_ds: DataSource,
-        to_ds: DataSource,
-        mapping: Mapping,
+        self, from_ds: DataSource, to_ds: DataSource, mapping: Mapping,
     ):
         operations = []
         from_class, to_class = mapping.entity_types
@@ -348,6 +379,12 @@ class Syncer:
 
     def _sync_entity(self, from_ds, from_entity, to_ds, to_class, mapping):
         operations = []
+
+        if mapping.dirty_mode == DirtyMode.UPDATE and not self._is_dirty(
+            mapping, from_entity
+        ):
+            return operations
+
         others = self.link_provider.others(from_entity, from_ds, to_class, to_ds)
         if not others or not len(others):
             operations.extend(
@@ -358,12 +395,7 @@ class Syncer:
         else:
             operations.extend(
                 self._update_linked_entity(
-                    from_ds,
-                    from_entity,
-                    others,
-                    mapping,
-                    to_class,
-                    to_ds,
+                    from_ds, from_entity, others, mapping, to_class, to_ds,
                 )
             )
         return operations
@@ -460,6 +492,8 @@ class Syncer:
             if len(key) > 1:
                 next_key, *rest = key
                 next_entity = getattr(entity, next_key)
+                if not next_entity:
+                    return None
                 return Syncer._resolve_key_value(next_entity, rest)
             elif len(key) == 1:
                 key = key[0]
